
#include <WiFi.h>
#include <WiFiClient.h>
#include <ESPmDNS.h>
#include <FS.h>
#include <SD.h>
#include <Wire.h>
#include "TimelyTaskManager.h"
#include "Storage.h"
#include "LCDAQM.h"
#include "INA219I2C.h"
#include "WebPageHandler.h"

//---------------------------------
// Constant

const String wifiSSID = "yourssid";
const String wifiPass = "yourpassword";
//#define staticIP 192,168,2,26           // Uncomment if use static IP

// Battery control
float vBatTarget = 14.0;          // Max target battery charge voltage
const float vBatG1 = 0.2;         // Adjust PWM if battery voltage get out of target +/- vBatG1
const float vBatG2 = 0.8;         // Fast adjust PWM if battery voltage get out of target +/- vBatG2
float cBatTarget = 5.0;           // Max target battery charge current (solar current - load current)
const float cBatG1 = 0.1;         // Adjust PWM if battery current get out of target +/- cBatG1
const float cBatG2 = 0.5;         // Fast adjust PWM if battery current get out of target +/- cBatG2
float vLoadCut = 10.8;            // Cut off load if battery voltage is lower than vLoadCut
float vLoadRecover = 12.0;        // Recover from cut off if battery voltage is higher than vLoadRecover

// Pin number
const uint8_t pinSW2 = 0;
const uint8_t pinSW3 = 4;
const uint8_t pinSW4 = 16;
const uint8_t pinSW5 = 5;
const uint8_t pinSDCS = 17;
const uint8_t pinSDA = 21;
const uint8_t pinSCL = 22;
const uint8_t pinPWM = 32;
const uint8_t pinLE = 13;
const uint8_t pinLoad = 33;
const uint8_t pinLED1 = 25;
const uint8_t pinLED2 = 26;
const uint8_t pinLED3 = 27;
const uint8_t pinLED4 = 14;
const uint8_t pinADC = 36;


// Menu
const uint8_t menuBatteryVC = 0;
const uint8_t menuBatteryP = 1;
const uint8_t menuChargerV = 2;
const uint8_t menuLog = 3;
const uint8_t menuLoad = 4;
const uint8_t menuSleep = 5;
const uint8_t lastMenuId = 5;


// Function ID for TimelyTaskManager
//  #9 is reserved for WebPageHandler
const uint8_t idDisplayUpdate = 0;
const uint8_t idSleepControl = 1;
const uint8_t idLoadControl = 2;
const uint8_t idLog = 3;


//---------------------------------
// Variable

TimelyTaskManager timelyTaskManager(10);  // Time management and scheduler
INA219I2C ina219(pinSDA, pinSCL);         // INA219
LCDAQM lcd(pinSDA, pinSCL);               // LCD
AsyncWebServer server(80);                // Web Server

uint16_t pwm = 0;                         // PWM duty
RTC_DATA_ATTR bool load = true;           // true:auto false:off
RTC_DATA_ATTR bool loadCut = false;       // ture:load cut and waiting for recovery
RTC_DATA_ATTR bool sleepMode = true;      // true:auto false:off
RTC_DATA_ATTR bool logMode = false;       // true:logging

// LED and Load state
//  Call changeState() to apply on devices
RTC_DATA_ATTR uint8_t stateLED1 = HIGH;
RTC_DATA_ATTR uint8_t stateLED2 = HIGH;
RTC_DATA_ATTR uint8_t stateLED3 = HIGH;
RTC_DATA_ATTR uint8_t stateLED4 = HIGH;
RTC_DATA_ATTR uint8_t stateLoad = HIGH;

bool wakeUpTimer = false;                 // Set true if wake up from deep sleep by timer
uint8_t menu = menuBatteryVC;             // LCD current menu
bool changeMode = false;                  // True when enter set mode in menu
bool tmpSetting = false;                  // Temporary setting in change mode

// Switch status for switch evenet
uint8_t prevSW2 = 1;
uint8_t prevSW3 = 1;
uint8_t prevSW4 = 1;
uint8_t prevSW5 = 1;

// Logging
RTC_DATA_ATTR char logPath[128] = "/default.log";   // Log file path
String logDir = "/log";                             // Log file directory


//---------------------------------
// Function

void setup() {
  Serial.begin(115200);
  delay(10);

  
  // Pin
  digitalWrite(pinLE, LOW);     // LOW = Disable
  digitalWrite(pinLoad, HIGH);  // HIGH = OFF
  digitalWrite(pinLED1, HIGH);  // HIGH = OFF
  digitalWrite(pinLED2, HIGH);  // HIGH = OFF
  digitalWrite(pinLED3, HIGH);  // HIGH = OFF
  digitalWrite(pinLED4, HIGH);  // HIGH = OFF
  pinMode(pinLE, OUTPUT);
  pinMode(pinLoad, OUTPUT);
  pinMode(pinLED1, OUTPUT);
  pinMode(pinLED2, OUTPUT);
  pinMode(pinLED3, OUTPUT);
  pinMode(pinLED4, OUTPUT);
  pinMode(pinSW2, PULLUP);
  pinMode(pinSW3, PULLUP);
  pinMode(pinSW4, PULLUP);
  pinMode(pinSW5, PULLUP);
  ledcSetup(0, 3000, 8);
  ledcAttachPin(pinPWM, 0);
  

  // Wake up reason
  esp_sleep_wakeup_cause_t wakeup_reason;
  wakeup_reason = esp_sleep_get_wakeup_cause();

  switch(wakeup_reason){
  case 1: // External signal RTC IO
    lcd.clear();
    sleepMode = false;  // Disable sleep mode if wake up by SW3
    ina219.vbus(true);
    ina219.current(true);
    loadControl();
    sleepControl();
    wifiControl(false);
  case 2: // External signal RTC CNTL
    break;
  case 3: // Timer
    wakeUpTimer = true;
    ina219.vbus(true);
    ina219.current(true);
    lcd.clear();
    if(logMode){
      oneTimeLog();
    }
    loadControl();
    sleepControl();
    wifiControl(false);
    break;
  case 4: // Touchpad
    break;
  case 5: // ULP program
    break;
  default: // Normal power on or reset
    powerOnInit();
    btStop();
    wifiControl(true);
    loadControl();
    sleepControl();
    break;
  }

  // TimelyTaskManager
  timelyTaskManager.setSecondlyTask(idDisplayUpdate, 1, displayUpdate);
  timelyTaskManager.setMinutelyTask(idSleepControl, 5, sleepControl);
  timelyTaskManager.setMinutelyTask(idLoadControl, 10, loadControl);
  if(logMode){
    timelyTaskManager.setMinutelyTask(idLog, 0, oneTimeLog);
  }

  // Show instructions on LCD
  lcd.clear();
  lcd.printStr("2:PGMDL");
  lcd.secLine();
  lcd.printStr("3:SET/WU");
  delay(2000);
  lcd.clear();
  lcd.printStr("4:+");
  lcd.secLine();
  lcd.printStr("5:-");
  delay(2000);

  menu = lastMenuId;
  swEvent(2);
}


void loop() {
  timelyTaskManager.updateTask();   // Handle scheduled tasks
  pwmControl();                     // Adjust charger PWM
  swEvent(swRead());                // Handle switch ON/OFF events
  delay(10);
}


// Initialize routine for power on or reset
void powerOnInit(){  
  // Disable sleep if SW3 is ON
  if(digitalRead(pinSW3)==0){
    sleepMode = false;
  }
  
  Storage::opLog("\n\n-----------------------\n");
  Storage::opLog("Program start\n");

  // Initialize LCD
  lcd.init();

  if(!sleepMode){
    lcd.printStr("SLEEP");
    lcd.secLine();
    lcd.printStr("DISABLED");
    delay(2000);
    lcd.clear();
  }
  
  // Initialize INA219
  if(!ina219.init()){
    Storage::opLog("INA219 Error\n");
    lcd.printStr("INA219");
    lcd.secLine();
    lcd.printStr("ERROR");

    // Stop program if error
    while(true){
      delay(1000);
    }
  }
}


// Call back from web page handler
void webPageHandlerCallBack(uint8_t action){
  // Not used in this application
  //switch(action){
  //  ...
  //}
}

// Change LED and load state through latch IC
void changeState(){
  digitalWrite(pinLED1, stateLED1);
  digitalWrite(pinLED2, stateLED2);
  digitalWrite(pinLED3, stateLED3);
  digitalWrite(pinLED4, stateLED4);
  digitalWrite(pinLoad, stateLoad);
  digitalWrite(pinLE, HIGH);
  delayMicroseconds(50);
  digitalWrite(pinLE, LOW);
}


// Return charger voltage from ESP-WROOM32 ADC
float vcharge(){
  float vcrg = 3.3 * 115 / 15 * analogRead(pinADC) / 4095;
  if(vcrg < 0.1){
    return 0;
  }
  return vcrg + 1.25; // 1.25V adjustment needed
}


// Return charger voltage by string
String vchargeStr(){
  char c[8];
  dtostrf(vcharge(), 3, 2, c);
  return String(c);
}


// Control PWM per battery voltage and current
void pwmControl(){
  float vBat = ina219.vbus(true);
  float cBat = ina219.current(true);
  float vCharge = vcharge();

  // Turn OFF if charger voltage is low
  if(vCharge < 10){
    pwm = 0;
    ledcWrite(0, pwm);
    return;
  }

  if(vBat > vBatTarget + vBatG2 || cBat > cBatTarget + cBatG2){
    if(pwm<=10){
      pwm = 0;
    }else{
      pwm-=10;
    }
  }else if(vBat > vBatTarget + vBatG1 || cBat > cBatTarget + cBatG1){
    if(pwm>=1){
      pwm-=1;
    }
  }else if(vBat < vBatTarget - vBatG2 && cBat < cBatTarget - cBatG2){
    if(pwm>=246){
      pwm = 256;
    }else{
      pwm+=10;
    }
  }else if(vBat < vBatTarget - vBatG1 && cBat < cBatTarget - cBatG1){
    if(pwm<=255){
      pwm+=1;
    }
  }

  ledcWrite(0, pwm);
}


// Check charger voltage and control deep sleep
void sleepControl(){
  // Enter deep sleep if charger voltage is <8V
  // Return to normal mode if charge voltage is >10V
  float vCharge = vcharge();
  float vth = 8;
  if(wakeUpTimer){
    vth = 10;
  }
  if(vCharge < vth && sleepMode){
    // Green LED OFF & Yellow LED ON
    stateLED1 = HIGH;
    stateLED2 = LOW;
    changeState();
    
    lcd.clear();
    lcd.printStr(ina219.vbusStr(false).c_str());
    lcd.printStr("V");
    lcd.secLine();
    // Only 1time log
    if(!wakeUpTimer){
      Storage::opLog(timelyTaskManager.getTimeStr() + " Low VCHARGE. Start deep sleep\n");
    }
    lcd.printStr("SLEEP");
    esp_sleep_enable_ext0_wakeup(GPIO_NUM_4, 0);  // Wake up if SW3 is pushed
    ESP.deepSleep(900000000);  // 15min deep sleep
  }
  
  // Green LED ON & Yellow LED OFF
  stateLED1 = LOW;
  stateLED2 = HIGH;
  changeState();

  wakeUpTimer = false;
}


// Check battery voltage and control load
void loadControl(){
  if(load){ // Auto
    float vBat = ina219.vbus(false);
    if(loadCut){
      if(vBat > vLoadRecover){
        stateLoad = LOW;  // Load on
        stateLED3 = HIGH;
        loadCut = false;
        Storage::opLog(timelyTaskManager.getTimeStr() + " VBATTERY recoverd. Connected the load.\n");
      }else{
        stateLoad = HIGH;  // Load off
        stateLED3 = LOW;
      }
    }else{
      if(vBat < vLoadCut){
        stateLoad = HIGH;  // Load off
        stateLED3 = LOW;
        loadCut = true;
        Storage::opLog(timelyTaskManager.getTimeStr() + " Low VBATTERY. Cut off the load.\n");
      }else{
        stateLoad = LOW;  // Load on
        stateLED3 = HIGH;
      }
    }
  }else{    // Off
    stateLoad = HIGH; // Load off
    stateLED3 = HIGH;
  }
  changeState();
}


// Change WiFi state
//  powerOnInit : true if power on or reset routine
void wifiControl(bool powerOnInit){
  if(powerOnInit){
    Storage::opLog("\nWiFi");
  }else{
    Storage::opLog(timelyTaskManager.getTimeStr() + " WiFi");
  }
  lcd.printStr("WIFI..");
  lcd.secLine();
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(wifiSSID.c_str(), wifiPass.c_str());

  // Wait WiFi Timeout 10sec
  for(uint16_t i=0; i<20; i++){
    if(WiFi.status()==WL_CONNECTED){
      #ifdef staticIP
      WiFi.config(IPAddress(staticIP), WiFi.gatewayIP(), WiFi.subnetMask(), WiFi.dnsIP());   // Set static IP address
      #endif
      
      Storage::opLog(" Connected IP : " + WiFi.localIP().toString() + "\n");
      lcd.printStr("CONNECTD");
      delay(1000);
      break;
    }else if(i==9){
      WiFi.begin(wifiSSID.c_str(), wifiPass.c_str()); // Sometimes 2nd time begin is needed
      delay(1000);
    }else if(i==19){
      Storage::opLog(" Not Found\n");
      lcd.printStr("NOTFOUND");
      WiFi.mode(WIFI_OFF);
      delay(1000);
      break;
    }
    delay(500);
  }

  // TimelyTaskManager (Need WiFi to get local time)
  timelyTaskManager.updateTime();
  Storage::opLog(timelyTaskManager.getTimeStr() + " TimelyTaskManager Started\n");

  // Web Server
  WebPageHandler::startServer(webPageHandlerCallBack);
  Storage::opLog(timelyTaskManager.getTimeStr() + " Web server started\n");
}


// Append battery status to log file one time
void oneTimeLog(){
  String timeStr = timelyTaskManager.getTimeStr(TimelyTaskManager::formatSecond);
  if(!Storage::exists(logDir)){
    Storage::mkdir(logDir);
  }
  bool exists = Storage::exists(logPath);
  File file = Storage::open(logPath, FILE_APPEND);
  if(file){
    if(!exists){
      file.print("Time,Battery V[V],Battery I[A],Battery P[W],Input V[V]\n");      
    }
    file.print(timeStr.c_str());
    file.print(",");
    file.print(ina219.vbusStr(false).c_str());
    file.print(",");
    file.print(ina219.currentStr(false).c_str());
    file.print(",");
    char p[8];
    dtostrf(ina219.vbus(false) * ina219.current(false), 1, 2, p);
    file.print(p);
    file.print(",");
    file.print(vchargeStr());
    file.print("\n");
    file.close();
  }
  Storage::end();
}

// Start logging
void startLog(){
  logMode = true;
  Storage::opLog(timelyTaskManager.getTimeStr() + " Start Log Mode\n");
  String s;
  if(logDir.endsWith("/")){
    s = logDir + timelyTaskManager.getTimeFileName(TimelyTaskManager::formatSecond) + ".csv";
  }else{
    s = logDir + "/" + timelyTaskManager.getTimeFileName(TimelyTaskManager::formatSecond) + ".csv";
  }
  strcpy(logPath, s.c_str());
  timelyTaskManager.setMinutelyTask(idLog, 0, oneTimeLog);
}

// Stop logging
void stopLog(){
  logMode = false;
  timelyTaskManager.removeTask(idLog);
  Storage::opLog(timelyTaskManager.getTimeStr() + " Finish Log Mode\n");
}

// Update display periodically
void displayUpdate(){
  switch(menu){
  case menuBatteryVC:
    displayBatteryVC();
    break;
  case menuBatteryP:
    displayBatteryP();
    break;
  case menuChargerV:
    displayChargeV();
    break;  
  case menuLog:
    lcd.clear();
    lcd.printStr("LOG");
    lcd.secLine();
    if(changeMode){
      if(tmpSetting){
        lcd.printStr("(ON)");
      }else{
        lcd.printStr("(OFF)");
      }
    }else{
      if(logMode){
        lcd.printStr("ON");
      }else{
        lcd.printStr("OFF");
      }
    }
    break;
  case menuLoad:
    lcd.clear();
    lcd.printStr("LOAD");
    lcd.secLine();
    if(changeMode){
      if(tmpSetting){
        lcd.printStr("(AUTO)");
      }else{
        lcd.printStr("(OFF)");
      }
    }else{
      if(load){
        lcd.printStr("AUTO");
      }else{
        lcd.printStr("OFF");
      }
    }
    break;
  case menuSleep:
    lcd.clear();
    lcd.printStr("SLEEP");
    lcd.secLine();
    if(changeMode){
      if(tmpSetting){
        lcd.printStr("(AUTO)");
      }else{
        lcd.printStr("(OFF)");
      }
    }else{
      if(sleepMode){
        lcd.printStr("AUTO");
      }else{
        lcd.printStr("OFF");
      }
    }
    break;
  }
}


// Display battery voltage and current on LCD
void displayBatteryVC(){  
  lcd.clear();
  lcd.printStr(ina219.vbusStr(false).c_str());
  lcd.printStr("V");
  lcd.secLine();
  lcd.printStr(ina219.currentStr(false).c_str());
  lcd.printStr("A");
}


// Display battery power on LCD
void displayBatteryP(){
  char p[8];
  dtostrf(ina219.vbus(false) * ina219.current(false), 1, 2, p);
  lcd.clear();
  lcd.printStr("BAT P");
  lcd.secLine();
  lcd.printStr(p);
  lcd.printStr("W");
}

// Display charger voltage on LCD
void displayChargeV(){
  lcd.clear();
  lcd.printStr("CRG V");
  lcd.secLine();
  lcd.printStr(vchargeStr().c_str());
  lcd.printStr("V");
}


// Monitor switches
//  Return pushed and released sw number
//  Return 0 if no switch change
uint8_t swRead(){
  // Switch2
  if(digitalRead(pinSW2)==0){
    prevSW2 = 0;
  }else{
    if(prevSW2 == 0){
      prevSW2 = 1;
      return 2;
    }
  }

  // Switch3
  if(digitalRead(pinSW3)==0){
    prevSW3 = 0;
  }else{
    if(prevSW3 == 0){
      prevSW3 = 1;
      return 3;
    }
  }
  
  // Switch4
  if(digitalRead(pinSW4)==0){
    prevSW4 = 0;
  }else{
    if(prevSW4 == 0){
      prevSW4 = 1;
      return 4;
    }
  }
  
  // Switch5
  if(digitalRead(pinSW5)==0){
    prevSW5 = 0;
  }else{
    if(prevSW5 == 0){
      prevSW5 = 1;
      return 5;
    }
  }
  return 0;
}


// Execute when SW is pushed
void swEvent(uint8_t sw){
  switch(sw){
  case 0:
    return;
    break;  // No event
    
  case 2:
    if(changeMode){return;}
    
    menu++;
    if(menu>lastMenuId){
      lcd.clear();
      lcd.printStr("BAT V");
      lcd.secLine();
      lcd.printStr("BAT C");
      delay(1000);
      menu = 0;
    }
    displayUpdate();
    break;  // Button2

  case 3:
    switch(menu){
    case menuBatteryVC:
      break;
    case menuBatteryP:
      break;
    case menuChargerV:
      break;
    case menuLog:
      if(changeMode){
        changeMode=false;
        if(tmpSetting && !logMode){
          startLog();
        }else if(!tmpSetting && logMode){
          stopLog();
        }
        displayUpdate();
      }else{
        changeMode=true;
        displayUpdate();
      }
      break;
      
    case menuLoad:
      if(changeMode){
        changeMode=false;
        load = tmpSetting;
        displayUpdate();
        loadControl();
      }else{
        changeMode=true;
        tmpSetting = load;
        displayUpdate();
      }
      break;
      
    case menuSleep:
      if(changeMode){
        changeMode=false;
        sleepMode = tmpSetting;
        displayUpdate();
        sleepControl();
      }else{
        changeMode=true;
        tmpSetting = sleepMode;
        displayUpdate();
      }
      break;
    }
    
    break;  // Button3

  case 4:
    if(!changeMode){
      menu++;
      if(menu>lastMenuId){
        lcd.clear();
        lcd.printStr("BAT V");
        lcd.secLine();
        lcd.printStr("BAT C");
        delay(1000);
        menu = 0;
      }
      displayUpdate();
    }else{
      switch(menu){
      case menuLog:
      case menuLoad:
      case menuSleep:
        if(changeMode){
          tmpSetting = !tmpSetting;
          displayUpdate();
        }
        break;
      }
    }
    break;  // Button4

  case 5:
    if(!changeMode){
      if(menu==1){
        lcd.clear();
        lcd.printStr("BAT V");
        lcd.secLine();
        lcd.printStr("BAT C");
        delay(1000);
        menu = 0;
      }else if(menu==0){
        menu = lastMenuId;
      }else{
        menu--;
      }
      displayUpdate();
    }else{
      switch(menu){
      case menuLog:
      case menuLoad:
      case menuSleep:
        if(changeMode){
          tmpSetting = !tmpSetting;
          displayUpdate();
        }
        break;
      }
    }
    break;  // Button5
  }
}
